(******************************************************************************

______________________________________________________________________________

PkYubikey                                                 (C) 2009, 2010 Pepak
http://www.pepak.net/bezpecnost/pkyubikey/                http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2010, Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uYubikey;
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

interface

uses
  libYubikey, Windows;

const
  YUBIKEY_ERROR_BASE = 12345678;
  YUBIKEY_ERROR_MAX = YUBIKEY_ERROR_BASE + 10;
  YUBIKEY_ERROR_INVALID_PARAMETERS = YUBIKEY_ERROR_BASE + 1;
  YUBIKEY_ERROR_UNKNOWN_NAME = YUBIKEY_ERROR_BASE + 2;
  YUBIKEY_ERROR_INVALID_OTP = YUBIKEY_ERROR_BASE + 3;
  YUBIKEY_ERROR_INVALID_DATA = YUBIKEY_ERROR_BASE + 4;

(* Register a new Yubikey to the system.
   Input:
       Name: PAnsiChar - identifier of the token (e.g. username)
       UID: PYubikeyUID - (secret) unique identifier of the token
       Key: PYubikeyKet - (secret) AES key of the token
       OtpCounter: integer - if larger than zero, current value of the OTP counter
   Output:
       Result: integer - error code of the operation. 0 means successful
                         registration, above codes mean specific Yubikey
                         errors, other values are Windows' GetLastError codes *)
function PKYubikeyRegisterOTP(Name: PAnsiChar; UID: PYubikeyUID; Key: PYubikeyKey; OtpCounter: Longint): Longint; stdcall;

(* Register a new Yubikey to the system. It accepts the same parameters and returns
   the same values as PKYoubikeyRegisterOTP, except that UID and Key are passed as
   strings (hex-encoded). *)
function PKYubikeyRegisterOTPhex(Name, UIDHex, KeyHex: PAnsiChar; OtpCounter: Longint): Longint; stdcall;

(* Authenticate a token's OTP:
   Input:
       Name: PAnsiChar - identifier of the token (e.g. username)
       OTP: PAnsiChar - one-time-password gathered from the Yubikey
   Output:
       Result: integer - error code of the operation. 0 means successful
                         authentication, above codes mean specific Yubikey
                         errors, other values are Windows' GetLastError codes *)
function PKYubikeyValidateOTP(Name, OTP: PAnsiChar): Longint; stdcall;

implementation

uses
  {$IFNDEF MINIMIZESIZE} SysUtils, {$ENDIF}
  {$IFDEF DEBUG} uDebug, {$ENDIF}
  uAnsiStrings;

const
  YUBIKEY_REG_HKEY = HKEY_LOCAL_MACHINE;
  YUBIKEY_REG_ROOTKEY = 'Software\Pepak\Yubikey\'; // Must end with backslash
  YUBIKEY_REG_UID = 'UID';
  YUBIKEY_REG_KEY = 'Key';
  YUBIKEY_REG_COUNTER = 'Counter';
  YUBIKEY_MUTEX = 'PkYubikey.';
  YUBIKEY_MUTEX_TIME = 1000; // Milliseconds

function PKYubikeyRegisterOTPhex(Name, UIDHex, KeyHex: PAnsiChar; OtpCounter: Longint): Longint;
var UID: TYubikeyUID;
    Key: TYubikeyKey;
begin
  YubikeyHexDecode(@UID, UIDHex, Sizeof(UID));
  YubikeyHexDecode(@Key, KeyHex, Sizeof(Key));
  Result := PKYubikeyRegisterOTP(Name, @UID, @Key, OtpCounter);
end;

const YUBIKEY_ROOTKEY_LENGTH = Cardinal(Length(YUBIKEY_REG_ROOTKEY));
      YUBIKEY_MUTEX_LENGTH = Cardinal(Length(YUBIKEY_MUTEX));

function OpenRegistryKey(Name: PAnsiChar; Create: boolean; out Mutex: THandle; out OwnMutex: boolean; out Reg: HKey): integer;
var MutexName, RegKeyName: PAnsiChar;
    NameLen: Cardinal;
begin
  {$IFDEF DEBUG} Debug('- OpenRegistryKey(''%s'',%d)', [Name, Integer(Create)]); {$ENDIF}
  NameLen := StrLen(Name);
  MutexName := StrAlloc(YUBIKEY_MUTEX_LENGTH + NameLen + 1);
  try
    StrCopy(MutexName, YUBIKEY_MUTEX);
    StrCat(MutexName, Name);
    Reg := 0;
    {$IFDEF DEBUG} Debug('  - Creating Mutex ''%s''', [MutexName]); {$ENDIF}
    Mutex := CreateMutex(nil, False, MutexName);
    if Mutex <> 0 then
      begin
      {$IFDEF DEBUG} Debug('  - Waiting until mutex is free'); {$ENDIF}
      Result := WaitForSingleObject(Mutex, YUBIKEY_MUTEX_TIME);
      if Result = WAIT_OBJECT_0 then
        begin
        OwnMutex := True;
        RegKeyName := StrAlloc(YUBIKEY_ROOTKEY_LENGTH + NameLen + 1);
        try
          StrCopy(RegKeyName, YUBIKEY_REG_ROOTKEY);
          StrCat(RegKeyName, Name);
          if Create then
            begin
            {$IFDEF DEBUG} Debug('  - Creating registry key ''%s''', [RegKeyName]); {$ENDIF}
            Result := RegCreateKeyEx(YUBIKEY_REG_HKEY, RegKeyName, 0, nil, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, nil, Reg, nil);
            end
          else
            begin
            {$IFDEF DEBUG} Debug('  - Opening registry key ''%s''', [RegKeyName]); {$ENDIF}
            Result := RegOpenKeyEx(YUBIKEY_REG_HKEY, RegKeyName, 0, KEY_ALL_ACCESS, Reg);
            end;
          if Result = ERROR_SUCCESS then
            begin
            Result := 0;
            end
          else
            begin
            Reg := 0;
            end;
        finally
          StrDispose(RegKeyName);
          end;
        end
      else
        OwnMutex := False;
      end
    else
      Result := GetLastError;
  finally
    StrDispose(MutexName);
    end;
  {$IFDEF DEBUG} Debug('  - Done with result code %d', [Result]); {$ENDIF}
end;

function CloseRegistryKey(Mutex: THandle; OwnMutex: boolean; Reg: HKey): integer;
begin
  {$IFDEF DEBUG} Debug('  - CloseRegistryKey(%d,%d,%d)', [Mutex, Integer(OwnMutex), Integer(Reg)]); {$ENDIF}
  if Reg <> 0 then
    RegCloseKey(Reg);
  if OwnMutex then
    ReleaseMutex(Mutex);
  if Mutex <> 0 then
    CloseHandle(Mutex);
  Result := 0;
  {$IFDEF DEBUG} Debug('  - Done with result code %d', [Result]); {$ENDIF}
end;

function PKYubikeyRegisterOTP(Name: PAnsiChar; UID: PYubikeyUID; Key: PYubikeyKey; OtpCounter: Longint): Longint;
var Mutex: THandle;
    OwnMutex: boolean;
    Reg: HKey;
begin
  {$IFDEF DEBUG} Debug(''); Debug('PKYubikeyRegisterOTP(''%s'', [%p], [%p], %d)', [Name, UID, Key, OtpCounter]); {$ENDIF}
  if (Name = nil) or (Key = nil) then
    begin
    Result := YUBIKEY_ERROR_INVALID_PARAMETERS;
    {$IFDEF DEBUG} Debug('PkYubikeyRegisterOTP FAILED: Invalid parameters'); {$ENDIF}
    end
  else
    begin
    Result := OpenRegistryKey(Name, True, Mutex, OwnMutex, Reg);
    try
      if Result = 0 then
        begin
        {$IFDEF DEBUG} Debug('- Setting registry'); {$ENDIF}
        if (RegSetValueEx(Reg, YUBIKEY_REG_UID, 0, REG_BINARY, UID, Sizeof(UID^)) = ERROR_SUCCESS) and
           (RegSetValueEx(Reg, YUBIKEY_REG_KEY, 0, REG_BINARY, Key, SizeOf(Key^)) = ERROR_SUCCESS) and
           ((OtpCounter < 0) or (RegSetValueEx(Reg, YUBIKEY_REG_COUNTER, 0, REG_DWORD, @OtpCounter, Sizeof(OtpCounter)) = ERROR_SUCCESS))
        then
          begin
          Result := 0;
          {$IFDEF DEBUG} Debug('- OK'); {$ENDIF}
          end
        else
          begin
          Result := GetLastError;
          {$IFDEF DEBUG} Debug('- OpenRegistryKey FAILED: Error code %d', [Result]); {$ENDIF}
          end;
        end
      else
        begin
        {$IFDEF DEBUG} Debug('- OpenRegistryKey FAILED: Error code %d', [Result]); {$ENDIF}
        end;
    finally
      CloseRegistryKey(Mutex, OwnMutex, Reg);
      end;
    end;
  {$IFDEF DEBUG} Debug('- Done with result code %d', [Result]); {$ENDIF}
end;

function PKYubikeyValidateOTP(Name, OTP: PAnsiChar): Longint;
var Mutex: THandle;
    OwnMutex: boolean;
    Reg: HKey;
    UID: TYubikeyUID;
    Key: TYubikeyKey;
    Token: TYubikeyToken;
    UIDSize, KeySize, RegCounterSize, OtpCounter, RegCounter: DWORD;
    b: boolean;
    i: integer;
begin
  {$IFDEF DEBUG} Debug(''); Debug('PKYubikeyValidateOTP(''%s'',''%s'')', [Name, OTP]); {$ENDIF}
  if (Name = nil) or (OTP = nil) or (StrLen(OTP) < 32) then
    begin
    Result := YUBIKEY_ERROR_INVALID_PARAMETERS;
    {$IFDEF DEBUG} Debug('PkYubikeyValidateOTP FAILED: Invalid parameters'); {$ENDIF}
    end
  else
    begin
    Result := OpenRegistryKey(Name, False, Mutex, OwnMutex, Reg);
    try
      if Result = 0 then
        begin
        UIDSize := Sizeof(UID);
        KeySize := Sizeof(Key);
        {$IFDEF DEBUG} Debug('- Reading UID and KEY from registry'); {$ENDIF}
        if (RegQueryValueEx(Reg, YUBIKEY_REG_UID, nil, nil, @UID, @UIDSize) = ERROR_SUCCESS) and
           (RegQueryValueEx(Reg, YUBIKEY_REG_KEY, nil, nil, @Key, @KeySize) = ERROR_SUCCESS)
        then
          begin
          {$IFDEF DEBUG} Debug('- Successful reading'); {$ENDIF}
          if (UIDSize = Sizeof(UID)) and (KeySize = Sizeof(Key)) then
            begin
            Result := YUBIKEY_ERROR_INVALID_OTP;
            // User's UID and AesKey have been found. Now I need to decrypt the OTP
            {$IFDEF DEBUG} Debug('- Parsing Yubikey OTP'); {$ENDIF}
            YubikeyParse(Pointer(OTP), @Key[0], @Token);
            // First check: Checksum must match
            if YubikeyCRC16(@Token, Sizeof(Token)) = YUBIKEY_CRC_OK_RESIDUE then
              begin
              {$IFDEF DEBUG} Debug('  - CRC16 OK. Comparing UID'); {$ENDIF}
              // Second check: UID must match
              b := True;
              for i := 0 to Pred(Sizeof(UID)) do
                if UID[i] <> Token.uid[i] then
                  begin
                  {$IFDEF DEBUG} Debug('  - UID mismatch at index %d: Expected %.2x, found %.2x', [i, Integer(UID[i]), Integer(Token.uid[i])]); {$ENDIF}
                  b := False;
                  Break;
                  end;
              if b then
                begin
                // Third check: Decrypted counter must be larger than stored counter
                {$IFDEF DEBUG} Debug('  - Reading COUNTER from registry'); {$ENDIF}
                RegCounterSize := Sizeof(RegCounter);
                if (RegQueryValueEx(Reg, YUBIKEY_REG_COUNTER, nil, nil, @RegCounter, @RegCounterSize) <> ERROR_SUCCESS) then
                  begin
                  RegCounter := 0;
                  {$IFDEF DEBUG} Debug('  - Failed. Assumed COUNTER = 0'); {$ENDIF}
                  end
                else if RegCounterSize <> Sizeof(RegCounter) then
                  begin
                  RegCounter := 0;
                  {$IFDEF DEBUG} Debug('  - Failed, read only %d bytes. Assumed COUNTER = 0', [RegCounterSize]); {$ENDIF}
                  end;
                OtpCounter := Token.ctr and $7fff;
                OtpCounter := OtpCounter shl 16 + Token.use;
                {$IFDEF DEBUG} Debug('  - Comparing registry counter (%d) to OTP counter (%d)', [RegCounter, OtpCounter]); {$ENDIF}
                if OtpCounter > RegCounter then
                  begin
                  // It's a valid OTP. Result code is set accordingly and the new Counter value is stored into registry
                  Result := 0;
                  {$IFDEF DEBUG} Debug('  - Storing new counter to registry'); {$ENDIF}
                  if RegSetValueEx(Reg, YUBIKEY_REG_COUNTER, 0, REG_DWORD, @OtpCounter, SizeOf(OtpCounter)) <> ERROR_SUCCESS then
                    begin
                    {$IFDEF DEBUG} Debug('  - FAILED write to registry'); {$ENDIF}
                    end;
                  end
                else
                  begin
                  {$IFDEF DEBUG} Debug('  - FAILED: OTP counter too small'); {$ENDIF}
                  end;
                end;
              end
            else
              begin
              {$IFDEF DEBUG} Debug('  - FAILED: CRC16 error'); {$ENDIF}
              end;
            end
          else
            begin
            Result := YUBIKEY_ERROR_INVALID_DATA;
            {$IFDEF DEBUG} Debug('- FAILED: Invalid data'); {$ENDIF}
            end;
          end
        else
          begin
          Result := YUBIKEY_ERROR_UNKNOWN_NAME;
          {$IFDEF DEBUG} Debug('- FAILED: Unknown name'); {$ENDIF}
          end;
        end
      else
        begin
        {$IFDEF DEBUG} Debug('- OpenRegistryKey FAILED: Error code %d', [Result]); {$ENDIF}
        end;
    finally
      CloseRegistryKey(Mutex, OwnMutex, Reg);
      end;
    end;
  {$IFDEF DEBUG} Debug('- Done with result code %d', [Result]); {$ENDIF}
end;

initialization

end.
